@@ -42,6 +42,7 @@ struct AsyncToGenerator;
 struct Actual {
     in_prototype_assignment: bool,
     extra_stmts: Vec<Stmt>,
+    hoist_stmts: Vec<Stmt>,
 }
 
 #[fast_path(ShouldWork)]
@@ -69,9 +70,11 @@ impl AsyncToGenerator {
             let mut actual = Actual {
                 in_prototype_assignment: false,
                 extra_stmts: vec![],
+                hoist_stmts: vec![],
             };
 
             stmt.visit_mut_with(&mut actual);
+            stmts_updated.extend(actual.hoist_stmts.into_iter().map(T::from_stmt));
             stmts_updated.push(stmt);
             stmts_updated.extend(actual.extra_stmts.into_iter().map(T::from_stmt));
         }
@@ -132,6 +135,7 @@ impl VisitMut for Actual {
                 ident: None,
                 function: m.function.take(),
             },
+            None,
             false,
         );
 
@@ -174,184 +178,7 @@ impl VisitMut for Actual {
     }
 
     fn visit_mut_expr(&mut self, expr: &mut Expr) {
-        match expr {
-            Expr::Paren(ParenExpr { expr, .. }) => {
-                expr.visit_mut_with(self);
-                return;
-            }
-
-            // Optimization for iife.
-            Expr::Call(CallExpr {
-                span,
-                callee: ExprOrSuper::Expr(callee),
-                args,
-                type_args,
-            }) if callee.is_fn_expr() => {
-                *expr = self.handle_iife(
-                    *span,
-                    &mut (*callee).take().expect_fn_expr(),
-                    args,
-                    type_args.take(),
-                );
-                return;
-            }
-
-            Expr::Call(CallExpr {
-                span,
-                callee: ExprOrSuper::Expr(callee),
-                args,
-                type_args,
-            }) if match &**callee {
-                Expr::Paren(ParenExpr { expr, .. }) => match &**expr {
-                    Expr::Fn(..) => true,
-                    _ => false,
-                },
-                _ => false,
-            } =>
-            {
-                *expr = self.handle_iife(
-                    *span,
-                    &mut (*callee).take().expect_paren().expr.expect_fn_expr(),
-                    args,
-                    type_args.take(),
-                );
-                return;
-            }
-
-            Expr::Assign(assign_expr) => {
-                // flag if expression is assignment to prototype chain should not try to bind
-                // this context
-                match &assign_expr.left {
-                    PatOrExpr::Pat(pat) => match &**pat {
-                        Pat::Expr(pat_expr) => match &**pat_expr {
-                            Expr::Member(MemberExpr {
-                                obj: ExprOrSuper::Expr(ex),
-                                computed: false,
-                                ..
-                            }) => match &**ex {
-                                Expr::Member(MemberExpr {
-                                    prop,
-                                    computed: false,
-                                    ..
-                                }) => match &**prop {
-                                    Expr::Ident(ident) => {
-                                        self.in_prototype_assignment = *ident.sym == *"prototype";
-                                    }
-                                    _ => {}
-                                },
-                                _ => {}
-                            },
-                            _ => {}
-                        },
-                        _ => {}
-                    },
-                    _ => {}
-                }
-
-                assign_expr.visit_mut_children_with(self);
-                self.in_prototype_assignment = false;
-                return;
-            }
-            _ => {}
-        }
-
-        expr.visit_mut_children_with(self);
-
-        match expr {
-            Expr::Arrow(ArrowExpr {
-                is_async: true,
-                span,
-                params,
-                body,
-                is_generator,
-                type_params,
-                return_type,
-            }) => {
-                // Apply arrow
-                let used_this = contains_this_expr(body);
-
-                let fn_expr = FnExpr {
-                    ident: None,
-                    function: Function {
-                        decorators: vec![],
-                        span: *span,
-                        params: params
-                            .into_iter()
-                            .map(|pat| Param {
-                                span: DUMMY_SP,
-                                decorators: Default::default(),
-                                pat: pat.take(),
-                            })
-                            .collect(),
-                        is_async: true,
-                        is_generator: *is_generator,
-                        body: Some(match body.take() {
-                            BlockStmtOrExpr::BlockStmt(block) => block,
-                            BlockStmtOrExpr::Expr(expr) => BlockStmt {
-                                span: DUMMY_SP,
-                                stmts: vec![Stmt::Return(ReturnStmt {
-                                    span: expr.span(),
-                                    arg: Some(expr),
-                                })],
-                            },
-                        }),
-                        type_params: type_params.take(),
-                        return_type: return_type.take(),
-                    },
-                };
-
-                if !used_this {
-                    *expr = make_fn_ref(fn_expr, false);
-                    return;
-                }
-
-                *expr = Expr::Call(CallExpr {
-                    span: span.take(),
-                    callee: make_fn_ref(fn_expr, false)
-                        .make_member(quote_ident!("bind"))
-                        .as_callee(),
-                    args: vec![ThisExpr { span: DUMMY_SP }.as_arg()],
-                    type_args: Default::default(),
-                });
-            }
-
-            Expr::Fn(
-                fn_expr @ FnExpr {
-                    function:
-                        Function {
-                            is_async: true,
-                            body: Some(..),
-                            ..
-                        },
-                    ..
-                },
-            ) => {
-                self.visit_fn(fn_expr.ident.clone(), &mut fn_expr.function, false);
-
-                let function = fn_expr.function.take();
-                let body = Some(BlockStmt {
-                    span: DUMMY_SP,
-                    stmts: self
-                        .extra_stmts
-                        .drain(..)
-                        .chain(function.body.unwrap().stmts)
-                        .collect(),
-                });
-
-                *expr = Expr::Call(CallExpr {
-                    span: DUMMY_SP,
-                    callee: Expr::Fn(FnExpr {
-                        ident: None,
-                        function: Function { body, ..function },
-                    })
-                    .as_callee(),
-                    args: vec![],
-                    type_args: Default::default(),
-                });
-            }
-
-            _ => {}
-        }
+        self.visit_mut_expr_with_binding(expr, None);
     }
 
     fn visit_mut_fn_decl(&mut self, f: &mut FnDecl) {
@@ -360,7 +187,7 @@ impl VisitMut for Actual {
             return;
         }
 
-        self.visit_fn(Some(f.ident.clone()), &mut f.function, true);
+        self.visit_fn(Some(f.ident.clone()), None, &mut f.function, true);
     }
 
     fn visit_mut_module_item(&mut self, item: &mut ModuleItem) {
@@ -369,7 +196,7 @@ impl VisitMut for Actual {
             ModuleItem::ModuleDecl(ModuleDecl::ExportDefaultDecl(export_default)) => {
                 if let DefaultDecl::Fn(expr) = &mut export_default.decl {
                     if expr.function.is_async {
-                        self.visit_fn(expr.ident.clone(), &mut expr.function, true);
+                        self.visit_fn(expr.ident.clone(), None, &mut expr.function, true);
                         *item = ModuleItem::ModuleDecl(ModuleDecl::ExportDefaultDecl(
                             ExportDefaultDecl {
                                 decl: DefaultDecl::Fn(expr.take()),
@@ -412,6 +239,7 @@ impl VisitMut for Actual {
                     ..prop.function.take()
                 },
             },
+            None,
             true,
         );
 
@@ -449,6 +277,34 @@ impl VisitMut for Actual {
     }
 
     fn visit_mut_stmts(&mut self, _n: &mut Vec<Stmt>) {}
+
+    fn visit_mut_var_declarator(&mut self, var: &mut VarDeclarator) {
+        if let VarDeclarator {
+            name: Pat::Ident(BindingIdent { id, .. }),
+            init: Some(init),
+            ..
+        } = var
+        {
+            match init.as_ref() {
+                Expr::Fn(FnExpr {
+                    ident: None,
+                    ref function,
+                }) if function.is_async || function.is_generator => {
+                    self.visit_mut_expr_with_binding(init, Some(id.clone()));
+                    return;
+                }
+
+                Expr::Arrow(arrow_expr) if arrow_expr.is_async || arrow_expr.is_generator => {
+                    self.visit_mut_expr_with_binding(init, Some(id.clone()));
+                    return;
+                }
+
+                _ => {}
+            }
+        }
+
+        var.visit_mut_children_with(self);
+    }
 }
 
 /// Hoists super access
@@ -737,6 +593,314 @@ impl VisitMut for MethodFolder {
 }
 
 impl Actual {
+    fn visit_mut_expr_with_binding(&mut self, expr: &mut Expr, binding_ident: Option<Ident>) {
+        match expr {
+            Expr::Paren(ParenExpr { expr, .. }) => {
+                expr.visit_mut_with(self);
+                return;
+            }
+
+            // Optimization for iife.
+            Expr::Call(CallExpr {
+                span,
+                callee: ExprOrSuper::Expr(callee),
+                args,
+                type_args,
+            }) if callee.is_fn_expr() => {
+                *expr = self.handle_iife(
+                    *span,
+                    &mut (*callee).take().expect_fn_expr(),
+                    args,
+                    type_args.take(),
+                );
+                return;
+            }
+
+            Expr::Call(CallExpr {
+                span,
+                callee: ExprOrSuper::Expr(callee),
+                args,
+                type_args,
+            }) if match &**callee {
+                Expr::Paren(ParenExpr { expr, .. }) => match &**expr {
+                    Expr::Fn(..) => true,
+                    _ => false,
+                },
+                _ => false,
+            } =>
+            {
+                *expr = self.handle_iife(
+                    *span,
+                    &mut (*callee).take().expect_paren().expr.expect_fn_expr(),
+                    args,
+                    type_args.take(),
+                );
+                return;
+            }
+
+            Expr::Assign(assign_expr) => {
+                // flag if expression is assignment to prototype chain should not try to bind
+                // this context
+                match &assign_expr.left {
+                    PatOrExpr::Pat(pat) => match &**pat {
+                        Pat::Expr(pat_expr) => match &**pat_expr {
+                            Expr::Member(MemberExpr {
+                                obj: ExprOrSuper::Expr(ex),
+                                computed: false,
+                                ..
+                            }) => match &**ex {
+                                Expr::Member(MemberExpr {
+                                    prop,
+                                    computed: false,
+                                    ..
+                                }) => match &**prop {
+                                    Expr::Ident(ident) => {
+                                        self.in_prototype_assignment = *ident.sym == *"prototype";
+                                    }
+                                    _ => {}
+                                },
+                                _ => {}
+                            },
+                            _ => {}
+                        },
+                        _ => {}
+                    },
+                    _ => {}
+                }
+
+                assign_expr.visit_mut_children_with(self);
+                self.in_prototype_assignment = false;
+                return;
+            }
+            _ => {}
+        }
+
+        expr.visit_mut_children_with(self);
+
+        match expr {
+            Expr::Arrow(ArrowExpr {
+                is_async: true,
+                span,
+                params,
+                body,
+                is_generator,
+                type_params,
+                return_type,
+            }) => {
+                // Apply arrow
+                let this: Option<ExprOrSpread> = if contains_this_expr(body) {
+                    if binding_ident.is_none() {
+                        Some(ThisExpr { span: DUMMY_SP }.as_arg())
+                    } else {
+                        let this = private_ident!("_this");
+                        self.hoist_stmts.push(Stmt::Decl(Decl::Var(VarDecl {
+                            span: DUMMY_SP,
+                            kind: VarDeclKind::Var,
+                            decls: vec![VarDeclarator {
+                                span: DUMMY_SP,
+                                name: this.clone().into(),
+                                init: Some(Box::new(ThisExpr { span: DUMMY_SP }.into())),
+                                definite: false,
+                            }],
+                            declare: false,
+                        })));
+                        Some(this.as_arg())
+                    }
+                } else {
+                    None
+                };
+
+                let fn_expr = FnExpr {
+                    ident: None,
+                    function: Function {
+                        decorators: vec![],
+                        span: *span,
+                        params: params
+                            .iter()
+                            .map(|pat| Param {
+                                span: DUMMY_SP,
+                                decorators: Default::default(),
+                                pat: pat.clone(),
+                            })
+                            .collect(),
+                        is_async: true,
+                        is_generator: *is_generator,
+                        body: Some(match body.take() {
+                            BlockStmtOrExpr::BlockStmt(block) => block,
+                            BlockStmtOrExpr::Expr(expr) => BlockStmt {
+                                span: DUMMY_SP,
+                                stmts: vec![Stmt::Return(ReturnStmt {
+                                    span: expr.span(),
+                                    arg: Some(expr),
+                                })],
+                            },
+                        }),
+                        type_params: type_params.take(),
+                        return_type: return_type.take(),
+                    },
+                };
+
+                let ref_ident = quote_ident!("ref");
+                let real_fn_ident = private_ident!(ref_ident.span, format!("_{}", ref_ident.sym));
+
+                let right = if let Some(this) = this {
+                    Expr::Call(CallExpr {
+                        span: *span,
+                        callee: make_fn_ref(fn_expr, Some(this.clone()), false)
+                            .make_member(quote_ident!("bind"))
+                            .as_callee(),
+                        args: vec![this.clone()],
+                        type_args: Default::default(),
+                    })
+                } else {
+                    make_fn_ref(fn_expr, None, false)
+                };
+
+                if binding_ident.is_none() {
+                    *expr = right;
+                    return;
+                }
+
+                let ref_stmt: Stmt = Stmt::Decl(
+                    VarDecl {
+                        span: DUMMY_SP,
+                        kind: VarDeclKind::Var,
+                        decls: vec![VarDeclarator {
+                            span: DUMMY_SP,
+                            name: Pat::Ident(real_fn_ident.clone().into()),
+                            init: Some(Box::new(right)),
+                            definite: false,
+                        }],
+                        declare: false,
+                    }
+                    .into(),
+                );
+
+                let apply = Stmt::Return(ReturnStmt {
+                    span: DUMMY_SP,
+                    arg: Some(Box::new(real_fn_ident.apply(
+                        DUMMY_SP,
+                        Box::new(Expr::This(ThisExpr { span: DUMMY_SP })),
+                        vec![quote_ident!("arguments").as_arg()],
+                    ))),
+                });
+
+                let function = Function {
+                    span: DUMMY_SP,
+                    is_async: false,
+                    is_generator: false,
+                    params: params
+                        .into_iter()
+                        .map_while(|pat| match pat {
+                            Pat::Ident(..) => Some(Param {
+                                span: DUMMY_SP,
+                                decorators: Default::default(),
+                                pat: pat.clone(),
+                            }),
+                            Pat::Array(..) | Pat::Object(..) => Some(Param {
+                                span: DUMMY_SP,
+                                decorators: Default::default(),
+                                pat: Pat::Ident(private_ident!("_").into()),
+                            }),
+                            _ => None,
+                        })
+                        .collect(),
+                    body: Some(BlockStmt {
+                        span: DUMMY_SP,
+                        stmts: vec![apply],
+                    }),
+                    decorators: Default::default(),
+                    type_params: Default::default(),
+                    return_type: Default::default(),
+                };
+
+                let return_function = Stmt::Return(ReturnStmt {
+                    span: DUMMY_SP,
+                    arg: Some(Box::new(
+                        FnExpr {
+                            function,
+                            ident: binding_ident,
+                        }
+                        .into(),
+                    )),
+                });
+
+                let block_stmt = BlockStmt {
+                    span: DUMMY_SP,
+                    stmts: vec![ref_stmt, return_function],
+                };
+
+                let function = Function {
+                    span: *span,
+                    body: Some(block_stmt),
+                    params: Default::default(),
+                    is_generator: false,
+                    is_async: false,
+                    decorators: Default::default(),
+                    return_type: Default::default(),
+                    type_params: Default::default(),
+                };
+
+                *expr = Expr::Call(CallExpr {
+                    span: DUMMY_SP,
+                    callee: Expr::Fn(FnExpr {
+                        ident: None,
+                        function,
+                    })
+                    .as_callee(),
+                    args: vec![],
+                    type_args: Default::default(),
+                });
+            }
+
+            Expr::Fn(
+                fn_expr @ FnExpr {
+                    function:
+                        Function {
+                            is_async: true,
+                            body: Some(..),
+                            ..
+                        },
+                    ..
+                },
+            ) => {
+                self.visit_fn(
+                    fn_expr.ident.clone(),
+                    binding_ident,
+                    &mut fn_expr.function,
+                    false,
+                );
+
+                let function = fn_expr.function.take();
+                let body = Some(BlockStmt {
+                    span: DUMMY_SP,
+                    stmts: self
+                        .extra_stmts
+                        .drain(..)
+                        .chain(function.body.unwrap().stmts)
+                        .collect(),
+                });
+
+                *expr = Expr::Call(CallExpr {
+                    span: DUMMY_SP,
+                    callee: Expr::Fn(FnExpr {
+                        ident: None,
+                        function: Function {
+                            body,
+                            params: Default::default(),
+                            ..function
+                        },
+                    })
+                    .as_callee(),
+                    args: vec![],
+                    type_args: Default::default(),
+                });
+            }
+
+            _ => {}
+        }
+    }
+
     fn handle_iife(
         &mut self,
         span: Span,
@@ -773,7 +937,7 @@ impl Actual {
             });
         }
 
-        let callee = make_fn_ref(callee.take(), self.in_prototype_assignment);
+        let callee = make_fn_ref(callee.take(), None, self.in_prototype_assignment);
 
         Expr::Call(CallExpr {
             span,
@@ -783,7 +947,13 @@ impl Actual {
         })
     }
 
-    fn visit_fn(&mut self, raw_ident: Option<Ident>, f: &mut Function, is_decl: bool) {
+    fn visit_fn(
+        &mut self,
+        function_ident: Option<Ident>,
+        binding_ident: Option<Ident>,
+        f: &mut Function,
+        is_decl: bool,
+    ) {
         if f.body.is_none() {
             return;
         }
@@ -808,9 +978,11 @@ impl Actual {
                         }
                     }
                 })
-                .collect()
+                .collect::<Vec<_>>()
         };
-        let ident = raw_ident.clone().unwrap_or_else(|| quote_ident!("ref"));
+        let ident = function_ident
+            .clone()
+            .unwrap_or_else(|| quote_ident!("ref"));
 
         let real_fn_ident = private_ident!(ident.span, format!("_{}", ident.sym));
         let right = make_fn_ref(
@@ -818,6 +990,7 @@ impl Actual {
                 ident: None,
                 function: f.take(),
             },
+            None,
             true,
         );
 
@@ -895,7 +1068,7 @@ impl Actual {
                         span: DUMMY_SP,
                         is_async: false,
                         is_generator: false,
-                        params: vec![],
+                        params: params.clone(),
                         body: Some(BlockStmt {
                             span: DUMMY_SP,
                             stmts: vec![apply],
@@ -905,25 +1078,23 @@ impl Actual {
                         return_type: Default::default(),
                     };
 
-                    if raw_ident.is_some() {
+                    if function_ident.is_some() {
                         vec![
                             Stmt::Decl(Decl::Fn(FnDecl {
-                                ident: raw_ident.clone().unwrap(),
+                                ident: function_ident.clone().unwrap(),
                                 declare: false,
                                 function: f,
                             })),
                             Stmt::Return(ReturnStmt {
                                 span: DUMMY_SP,
-                                arg: Some(Box::new(Expr::Ident(raw_ident.clone().unwrap()))),
+                                arg: Some(Box::new(Expr::Ident(function_ident.clone().unwrap()))),
                             }),
                         ]
                     } else {
+                        let ident = function_ident.or(binding_ident);
                         vec![Stmt::Return(ReturnStmt {
                             span: DUMMY_SP,
-                            arg: Some(Box::new(Expr::Fn(FnExpr {
-                                ident: raw_ident,
-                                function: f,
-                            }))),
+                            arg: Some(Box::new(Expr::Fn(FnExpr { ident, function: f }))),
                         })]
                     }
                 },
@@ -941,7 +1112,11 @@ impl Actual {
 /// Creates
 ///
 /// `_asyncToGenerator(function*() {})` from `async function() {}`;
-fn make_fn_ref(mut expr: FnExpr, should_not_bind_this: bool) -> Expr {
+fn make_fn_ref(
+    mut expr: FnExpr,
+    this_expr: Option<ExprOrSpread>,
+    should_not_bind_this: bool,
+) -> Expr {
     expr.function.body.visit_mut_with(&mut AsyncFnBodyHandler {
         is_async_generator: expr.function.is_generator,
     });
@@ -964,7 +1139,7 @@ fn make_fn_ref(mut expr: FnExpr, should_not_bind_this: bool) -> Expr {
         Expr::Call(CallExpr {
             span: DUMMY_SP,
             callee: expr.make_member(quote_ident!("bind")).as_callee(),
-            args: vec![ThisExpr { span: DUMMY_SP }.as_arg()],
+            args: vec![this_expr.unwrap_or(ThisExpr { span: DUMMY_SP }.as_arg())],
             type_args: Default::default(),
         })
     } else {
